import django, os
from djcelery_transactions import task, shared_task

if django.VERSION < (1, 8):
    from django.test import TransactionTestCase as TestCaseForTests  # commit() was disabled in TestCase prior to 1.8
else:
    from django.test import TestCase as TestCaseForTests  # Django 1.8 now works just fine with TestCase
from .models import Trees, Plants

if django.VERSION < (1, 7):
    from django.core.cache import cache
else:
    from django.core.cache import caches

    cache = caches['default']

if django.VERSION >= (1, 6):
    from django.db.transaction import atomic
else:
    from django.db import transaction

    atomic = transaction.commit_on_success


@task
def my_task():
    cache.set('my_global', 42)


@task
def my_model_task():
    Plants.objects.create(name='Oak')


@shared_task
def my_shared_task():
    cache.set('my_global', 42)


try:
    from celery.registry import tasks

    tasks.register(my_task)
except:
    from celery import task as base_task, current_app, Task

    current_app.registry.register(my_task)


class SpecificException(Exception):
    pass


class DjangoCeleryTestCase(TestCaseForTests):
    """Test djcelery transaction safe task manager
    """

    def setUp(self):
        super(DjangoCeleryTestCase, self).setUp()
        self.task = my_task

    def tearDown(self):
        cache.delete('my_global')

    def test_committed_transaction_fire_task(self):
        """Check that task is consumed when no exception happens
        """

        @atomic
        def do_something():
            self.task.delay()

        do_something()
        self.assertEqual(cache.get('my_global'), 42)

    def test_committed_nested_transaction_fire_task(self):
        """Check that task is consumed when no exception happens
        """

        @atomic
        def do_something():
            @atomic
            def nested_do_something():
                self.task.delay()

            nested_do_something()

        do_something()
        self.assertEqual(cache.get('my_global'), 42)

    def test_rollbacked_transaction_discard_task(self):
        """Check that task is not consumed when exception happens
        """

        @atomic
        def do_something():
            self.task.delay()
            raise SpecificException

        try:
            do_something()
        except SpecificException:
            self.assertIsNone(cache.get('my_global'))
        else:
            self.fail('Exception not raised')

    def test_via_api(self):

        r = self.client.get('/test_api/')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(cache.get('my_global'), 42)

    def test_django_db_transaction_managed(self):
        """
        Check that django.db.transaction.managed is not affected
        by monkey-patching
        """

        if django.VERSION >= (1, 6):
            self.skipTest('Django 1.6 does not need this test')

        from django.db import transaction
        self.assertFalse(transaction.is_managed())
        transaction.enter_transaction_management()
        try:
            transaction.managed()
            self.assertTrue(transaction.is_managed())
        finally:
            transaction.leave_transaction_management()

    def test_multiple_models(self):
        """Check that task is consumed when no exception happens
        """

        self.assertEqual(Plants.objects.count(), 0)

        @atomic
        def do_something():
            my_model_task.delay()

        do_something()
        self.assertEqual(Plants.objects.count(), 1)

        Trees.objects.create(name='Grey Oak', plant=Plants.objects.get(name='Oak'))

    def test_nested_atomic_blocks(self):
        """Check that task is consumed only after last transaction
        """

        @atomic
        def do_something():
            self.task.delay()
            self.assertIsNone(cache.get('my_global'))

        @atomic
        def do_something_outside():
            do_something()
            self.assertIsNone(cache.get('my_global'))

        do_something_outside()
        self.assertEqual(cache.get('my_global'), 42)


class SharedTaskTestCase(DjangoCeleryTestCase):
    """Test djcelery transaction safe task manager
    """

    def setUp(self):
        super(SharedTaskTestCase, self).setUp()
        self.task = my_shared_task



# coding=utf-8
from functools import partial
import threading
from celery import current_app

from celery import current_app, Task
from celery import task as base_task, shared_task as base_shared_task
from celery.contrib.batches import Batches
import django
from django.conf import settings
from django.db import transaction

from django.db.transaction import get_connection, atomic

import djcelery_transactions.transaction_signals

# Thread-local data (task queue).
_thread_data = threading.local()

def _get_task_queue():
    """Returns the calling thread's task queue."""
    return _thread_data.__dict__.setdefault("task_queue", [])


class PostTransactionTask(Task):
    """A task whose execution is delayed until after the current transaction.

    The task's fate depends on the outcome of the current transaction. If it's
    committed or no changes are made in the transaction block, the task is sent
    as normal. If it's rolled back, the task is discarded.

    If transactions aren't being managed when ``apply_asyc()`` is called (if
    you're in the Django shell, for example) or the ``after_transaction``
    keyword argument is ``False``, the task will
    A replacement decorator is provided:

    .. code-block:: python

        from djcelery_transactions import task

        @task
        def example(pk):
            print "Hooray, the transaction has been committed!"
    """

    abstract = True

    def original_apply_async(self, *args, **kwargs):
        """Shortcut method to reach real implementation
        of celery.Task.apply_sync
        """
        return super(PostTransactionTask, self).apply_async(*args, **kwargs)

    def apply_async(self, *args, **kwargs):
        # Delay the task unless the client requested otherwise or transactions
        # aren't being managed (i.e. the signal handlers won't send the task).

        celery_eager = _get_celery_settings('CELERY_ALWAYS_EAGER')

        # New setting to run eager task post transaction
        # defaults to `not CELERY_ALWAYS_EAGER`
        eager_transaction = _get_celery_settings('CELERY_EAGER_TRANSACTION',
                                                 not celery_eager)

        if django.VERSION < (1, 6):

            if transaction.is_managed() and eager_transaction:
                if not transaction.is_dirty():
                    # Always mark the transaction as dirty
                    # because we push task in queue that must be fired or discarded
                    if 'using' in kwargs:
                        transaction.set_dirty(using=kwargs['using'])
                    else:
                        transaction.set_dirty()
                _get_task_queue().append((self, args, kwargs))
            else:
                apply_async_orig = super(PostTransactionTask, self).apply_async
                return apply_async_orig(*args, **kwargs)

        else:

            connection = get_connection()
            if connection.in_atomic_block and eager_transaction:
                _get_task_queue().append((self, args, kwargs))
            else:
                return self.original_apply_async(*args, **kwargs)


class PostTransactionBatches(Batches):
    """A batch of tasks whose queuing is delayed until after the current
        transaction.
    """

    abstract = True

    def original_apply_async(self, *args, **kwargs):
        """Shortcut method to reach real implementation
        of celery.Task.apply_sync
        """
        return super(PostTransactionBatches, self).apply_async(*args, **kwargs)

    def apply_async(self, *args, **kwargs):
        # Delay the task unless the client requested otherwise or transactions
        # aren't being managed (i.e. the signal handlers won't send the task).

        celery_eager = _get_celery_settings('CELERY_ALWAYS_EAGER')

        # New setting to run eager task post transaction
        # defaults to `not CELERY_ALWAYS_EAGER`
        eager_transaction = _get_celery_settings('CELERY_EAGER_TRANSACTION',
                                                 not celery_eager)

        connection = get_connection()
        if connection.in_atomic_block and eager_transaction:
            _get_task_queue().append((self, args, kwargs))
        else:
            return self.original_apply_async(*args, **kwargs)

def _discard_tasks(**kwargs):
    """Discards all delayed Celery tasks.

    Called after a transaction is rolled back."""
    _get_task_queue()[:] = []


def _send_tasks(**kwargs):
    """Sends all delayed Celery tasks.

    Called after a transaction is committed or we leave a transaction
    management block in which no changes were made (effectively a commit).
    """

    # Detect test mode through CELERY_ALWAYS_EAGER settings
    # We assume all celery transactions tests on 1.8+ are running with TestCase, otherwise we'd get atomic exceptions
    celery_eager = _get_celery_settings('CELERY_ALWAYS_EAGER')

    # If we detect higher up nested atomic block, continue
    connection = get_connection()
    if django.VERSION >= (1, 8):
        min_number_transactions = 1 if celery_eager else 0
        if (not celery_eager and connection.in_atomic_block) or len(connection.savepoint_ids) > min_number_transactions:
            return
    elif connection.in_atomic_block:
        return

    queue = _get_task_queue()
    while queue:
        tsk, args, kwargs = queue.pop(0)
        tsk.original_apply_async(*args, **kwargs)


# Replacement decorators.
task = partial(base_task, base=PostTransactionTask)
shared_task = partial(base_shared_task, base=PostTransactionTask)

# Hook the signal handlers up.
transaction.signals.post_commit.connect(_send_tasks)
transaction.signals.post_rollback.connect(_discard_tasks)

def _get_celery_settings(setting, default=False):
    """ Returns CELERY setting
    :param setting:
    :param default:
    :return:
    """
    return any(getattr(obj, setting, default)
               for obj in (current_app.conf, settings))



# coding=utf-8
"""Adds signals to ``django.db.transaction``.

Signals are monkey patched into ``django.db.transaction``:

* ``post_commit``: sent after a transaction is committed. If no changes were
  made in the transaction block, nothing is committed and this won't be sent.
* ``post_rollback``: sent after a transaction is rolled back.
* ``post_transaction_management``: sent after leaving transaction management.
  This signal isn't  posted if a ``TransactionManagementError`` is raised.

.. code-block:: python

    import djcelery_transactions.transaction_signals


    def _post_commit(**kwargs):
        print "The transaction has been committed!"


    django.db.transaction.signals.post_commit.connect(_post_commit)

This code was inspired by Grégoire Cachet's implementation of similar
functionality, which can be found on GitHub: https://gist.github.com/247844

.. warning::

    This module must be imported before you attempt to use the signals.
"""
from functools import partial
import django

from django.db import (
    DEFAULT_DB_ALIAS, DatabaseError, ProgrammingError, Error, connections,
)

from django.dispatch import Signal
from django.db.transaction import get_connection

from django.db import transaction

class TransactionSignals(object):
    """A container for the transaction signals."""

    def __init__(self):
        self.post_commit = Signal()
        self.post_rollback = Signal()

# Access as django.db.transaction.signals.
transaction.signals = TransactionSignals()


__original__exit__ = transaction.Atomic.__exit__

def __patched__exit__(self, exc_type, exc_value, traceback):
    connection = get_connection(self.using)

    if connection.savepoint_ids:
        sid = connection.savepoint_ids.pop()
    else:
        # Prematurely unset this flag to allow using commit or rollback.
        connection.in_atomic_block = False

    try:
        if connection.closed_in_transaction:
            # The database will perform a rollback by itself.
            # Wait until we exit the outermost block.
            pass

        elif exc_type is None and not connection.needs_rollback:
            if connection.in_atomic_block:
                # Release savepoint if there is one
                if sid is not None:
                    try:
                        connection.savepoint_commit(sid)
                        transaction.signals.post_commit.send(None)
                    except DatabaseError:
                        try:
                            connection.savepoint_rollback(sid)
                            transaction.signals.post_rollback.send(None)
                            # The savepoint won't be reused. Release it to
                            # minimize overhead for the database server.
                            connection.savepoint_commit(sid)
                            transaction.signals.post_commit.send(None)
                        except Error:
                            # If rolling back to a savepoint fails, mark for
                            # rollback at a higher level and avoid shadowing
                            # the original exception.
                            connection.needs_rollback = True
                        raise
            else:
                # Commit transaction
                try:
                    connection.commit()
                    transaction.signals.post_commit.send(None)
                except DatabaseError:
                    try:
                        connection.rollback()
                        transaction.signals.post_rollback.send(None)
                    except Error:
                        # An error during rollback means that something
                        # went wrong with the connection. Drop it.
                        connection.close()
                    raise
        else:
            # This flag will be set to True again if there isn't a savepoint
            # allowing to perform the rollback at this level.
            connection.needs_rollback = False
            if connection.in_atomic_block:
                # Roll back to savepoint if there is one, mark for rollback
                # otherwise.
                if sid is None:
                    connection.needs_rollback = True
                else:
                    try:
                        connection.savepoint_rollback(sid)
                        transaction.signals.post_rollback.send(None)
                        # The savepoint won't be reused. Release it to
                        # minimize overhead for the database server.
                        connection.savepoint_commit(sid)
                        transaction.signals.post_commit.send(None)
                    except Error:
                        # If rolling back to a savepoint fails, mark for
                        # rollback at a higher level and avoid shadowing
                        # the original exception.
                        connection.needs_rollback = True
            else:
                # Roll back transaction
                try:
                    connection.rollback()
                    transaction.signals.post_rollback.send(None)
                except Error:
                    # An error during rollback means that something
                    # went wrong with the connection. Drop it.
                    connection.close()

    finally:
        # Outermost block exit when autocommit was enabled.
        if not connection.in_atomic_block:
            if connection.closed_in_transaction:
                connection.connection = None
            elif connection.features.autocommits_when_autocommit_is_off:
                connection.autocommit = True
            else:
                connection.set_autocommit(True)
        # Outermost block exit when autocommit was disabled.
        elif not connection.savepoint_ids and not connection.commit_on_exit:
            if connection.closed_in_transaction:
                connection.connection = None
            else:
                connection.in_atomic_block = False


# Monkey patch that shit
transaction.Atomic.__exit__ = __patched__exit__


# -*- coding: utf-8 -*-
import django
from django.conf import settings
from django.core.management import call_command

import os, sys
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")
sys.path.insert(0, 'tests')


def runtests():
    if not settings.configured:
        # Choose database for settings
        DATABASES = {
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:'
            }
        }
        test_db = os.environ.get('DB', 'sqlite')
        if test_db == 'mysql':
            DATABASES['default'].update({
                'ENGINE': 'django.db.backends.mysql',
                'NAME': 'testdb',
                'USER': 'root',
            })

        # Configure test environment
        settings.configure(
            DATABASES=DATABASES,
            INSTALLED_APPS=(
                'djcelery_transactions',
                'tests.test',
            ),
            ROOT_URLCONF='tests.urls',
            LANGUAGES=(
                ('en', 'English'),
            ),
            MIDDLEWARE_CLASSES=(),
            CELERY_EAGER_TRANSACTION = True,
            TEST_RUNNER = 'djcelery.contrib.test_runner.CeleryTestSuiteRunner'
        )

        from celery import current_app
        current_app.conf.CELERY_ALWAYS_EAGER = True
        current_app.conf.CELERY_EAGER_PROPAGATES_EXCEPTIONS = True

    if django.VERSION >= (1, 7):
        django.setup()
    failures = call_command(
        'test', 'tests', interactive=False, failfast=False, verbosity=1)

    sys.exit(bool(failures))


if __name__ == '__main__':
    runtests()



from django.db import models


class Plants(models.Model):
    name = models.CharField(max_length=120)


class Trees(models.Model):
    name = models.CharField(max_length=120)
    plant = models.ForeignKey(Plants)



from django.http import HttpResponse
from .tests import my_task
import django

if django.VERSION > (1,6):
    from django.db.transaction import atomic
else:
    from django.db import transaction
    atomic = transaction.commit_on_success

@atomic
def test_api(request):
    my_task.delay()
    return HttpResponse('ok')